#ifndef CVD_BRESENHAM_H
#define CVD_BRESENHAM_H

#include <cvd/brezenham.h>

namespace CVD
{
/// @ingroup gGraphics
/// Alternative (and correct) spelling for this object
typedef Brezenham Bresenham;
/// @ingroup gGraphics
/// Alternative (and correct) spelling for this object
typedef Brezenham8 Bresenham8;
}

#endif
